% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nas.R
\name{is_na}
\alias{is_na}
\alias{is_na.default}
\alias{is_na.POSIXlt}
\alias{is_na.vctrs_rcrd}
\alias{is_na.data.frame}
\alias{num_na}
\alias{which_na}
\alias{which_not_na}
\alias{any_na}
\alias{all_na}
\alias{row_na_counts}
\alias{col_na_counts}
\alias{row_all_na}
\alias{col_all_na}
\alias{row_any_na}
\alias{col_any_na}
\title{Efficient functions for dealing with missing values.}
\usage{
is_na(x)

\method{is_na}{default}(x)

\method{is_na}{POSIXlt}(x)

\method{is_na}{vctrs_rcrd}(x)

\method{is_na}{data.frame}(x)

num_na(x, recursive = TRUE)

which_na(x)

which_not_na(x)

any_na(x, recursive = TRUE)

all_na(x, recursive = TRUE)

row_na_counts(x, names = FALSE)

col_na_counts(x, names = FALSE)

row_all_na(x, names = FALSE)

col_all_na(x, names = FALSE)

row_any_na(x, names = FALSE)

col_any_na(x, names = FALSE)
}
\arguments{
\item{x}{A vector, list, data frame or matrix.}

\item{recursive}{Should the function be applied recursively to lists?
The default is \code{TRUE}. Setting this to \code{TRUE} is actually much cheaper because
when \code{FALSE}, the other \code{NA} functions rely on calling \code{is_na()},
therefore allocating a vector. This is so that alternative objects with
\code{is.na} methods can be supported.}

\item{names}{Should row/col names be added?}
}
\value{
Number or location of \code{NA} values.
}
\description{
\code{is_na()} is a parallelised alternative to \code{is.na()}. \cr
\code{num_na(x)} is a faster and more efficient \code{sum(is.na(x))}. \cr
\code{which_na(x)} is a more efficient \code{which(is.na(x))} \cr
\code{which_not_na(x)} is a more efficient \code{which(!is.na(x))} \cr
\code{row_na_counts(x)} is a more efficient \code{rowSums(is.na(x))} \cr
\code{row_all_na()} returns a logical vector indicating which rows are empty
and have only \code{NA} values. \cr
\code{row_any_na()} returns a logical vector indicating which rows have at least
1 \code{NA} value. \cr
The \code{col_} variants are the same, but operate by-column.
}
\details{
These functions are designed primarily for programmers, to increase the speed
and memory-efficiency of \code{NA} handling. \cr
Most of these functions can be parallelised through \code{options(cheapr.cores)}. \cr
\subsection{Common use-cases}{

To replicate \code{complete.cases(x)}, use \code{!row_any_na(x)}. \cr
To find rows with any empty values,
use \code{which_(row_any_na(df))}. \cr
To find empty rows use \code{which_(row_all_na(df))} or \code{which_na(df)}.
To drop empty rows use \code{na_rm(df)} or \code{sset(df, which_(row_all_na(df), TRUE))}.
}

\subsection{\code{is_na}}{

\code{is_na} Is an S3 generic function. It will internally fall back on
using \code{is.na} if it can't find a suitable method.
Alternatively you can write your own \code{is_na} method.
For example there is a method for \code{vctrs_rcrd}
objects that simply converts it to a data frame and then calls \code{row_all_na()}.
There is also a \code{POSIXlt} method for \code{is_na} that is much faster than \code{is.na}.
}

\subsection{Lists}{

When \code{x} is a list, \code{num_na}, \code{any_na} and \code{all_na} will recursively search
the list for \code{NA} values. If \code{recursive = F} then \code{is_na()} is used to
find \code{NA} values. \cr
\code{is_na} differs to \code{is.na} in 2 ways:
\itemize{
\item List elements are counted as \code{NA} if either that value is \code{NA}, or
if it's a list, then all values of that list are \code{NA}.
\item When called on a data frame, it returns \code{TRUE} for empty rows that contain
only \code{NA} values.
}
}
}
\examples{
library(cheapr)
library(bench)

x <- 1:10
x[c(1, 5, 10)] <- NA
num_na(x)
which_na(x)
which_not_na(x)

row_nas <- row_na_counts(airquality, names = TRUE)
col_nas <- col_na_counts(airquality, names = TRUE)
row_nas
col_nas

df <- sset(airquality, j = 1:2)

# Number of NAs in data
num_na(df)
# Which rows are empty?
row_na <- row_all_na(df)
sset(df, row_na)

# Removing the empty rows
sset(df, which_(row_na, invert = TRUE))
# Or
na_rm(df)
# Or
sset(df, row_na_counts(df) < ncol(df))
}
